#version 400 compatibility

/*
Copyright (C) 2019 RRe36

All Rights Reserved unless otherwise explicitly stated.


By downloading this you have agreed to the license and terms of use.
These can be found inside the included license-file or here: https://rre36.github.io/kappa_shader_web/license/

Violating these terms may be penalized with actions according to the Digital Millennium Copyright Act (DMCA), the Information Society Directive and/or similar laws depending on your country.
*/

#include "/lib/common.glsl"
#include "/lib/util/encoders.glsl"

const int shadowMapResolution   = 2560;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]

const int noiseTextureResolution = 256;

in vec2 coord;

flat in mat4x3 light_color;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex3;
uniform sampler2D colortex5;

uniform sampler2D depthtex1;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform int worldTime;
uniform int frameCounter;

uniform float wetness;

uniform vec4 daytime;
uniform vec2 viewSize;
uniform vec2 pixelSize;

uniform float aspectRatio;
uniform float eyeAltitude;
uniform float frameTimeCounter;
uniform float far, near;

uniform vec3 upvec, upvecView;
uniform vec3 sunvec, sunvecView;
uniform vec3 moonvec, moonvecView;
uniform vec3 lightvec;
uniform vec3 cloud_lvec, cloud_lvecView;

uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

vec3 screen_viewspace(vec3 screenpos, mat4 projInv) {
    screenpos   = screenpos*2.0-1.0;

    vec3 viewpos    = vec3(vec2(projInv[0].x, projInv[1].y)*screenpos.xy + projInv[3].xy, projInv[3].z);
        viewpos    /= projInv[2].w*screenpos.z + projInv[3].w;
    
    return viewpos;
}

vec3 screen_viewspace(vec3 screenpos) {
    return screen_viewspace(screenpos, gbufferProjectionInverse);
}

vec3 view_scenespace(vec3 viewpos, mat4 mvInv) {
    return viewMAD(mvInv, viewpos);
}

vec3 view_scenespace(vec3 viewpos) {
    return view_scenespace(viewpos, gbufferModelViewInverse);
}

float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

#include "/lib/atmos/phase.glsl"

#include "/lib/atmos/asky.glsl"

vec4 textureBicubic(sampler2D sampler, vec2 coord) {
	vec2 res = textureSize(sampler, 0);

	coord = coord * res - 0.5;

	vec2 f = fract(coord);
	coord -= f;

	vec2 ff = f * f;
	vec4 w0;
	vec4 w1;
	w0.xz = 1 - f; w0.xz *= w0.xz * w0.xz;
	w1.yw = ff * f;
	w1.xz = 3 * w1.yw + 4 - 6 * ff;
	w0.yw = 6 - w1.xz - w1.yw - w0.xz;

	vec4 s = w0 + w1;
	vec4 c = coord.xxyy + vec2(-0.5, 1.5).xyxy + w1 / s;
	c /= res.xxyy;

	vec2 m = s.xz / (s.xz + s.yw);
	return mix(
		mix(texture(sampler, c.yw), texture(sampler, c.xw), m.x),
		mix(texture(sampler, c.yz), texture(sampler, c.xz), m.x),
		m.y);
}

const float gauss9w[9] = float[9] (
     0.0779, 0.12325, 0.0779,
    0.12325, 0.1954,  0.12225,
     0.0779, 0.12325, 0.0779
);

const vec2 gauss9o[9] = vec2[9] (
    vec2(1.0, 1.0), vec2(0.0, 1.0), vec2(-1.0, 1.0),
    vec2(1.0, 0.0), vec2(0.0, 0.0), vec2(-1.0, 0.0),
    vec2(1.0, -1.0), vec2(0.0, -1.0), vec2(-1.0, -1.0)
);

vec4 gauss9(sampler2D tex, vec2 coord, float sigma) {
    vec2 pixelRad = sigma*pixelSize;
    vec4 col = vec4(0.0);

    for (int i = 0; i<9; i++) {
        vec2 bcoord = coord + gauss9o[i]*pixelRad;
        col += texture(tex, bcoord)*gauss9w[i];
    }
    return col;
}

#include "/lib/atmos/project.glsl"

vec3 get_sun(vec3 viewvec) {
    vec3 v      = -viewvec;
    vec3 sv     = normalize(sunvecView+v);
    float sun   = dot(sv, v);

    const float size = sun_size * 0.0055;
    float maxsize = size + 0.0004;
        maxsize  += lin_step(sunvec.y, -0.04, 0.04)*0.004;

    float s   = 1.0-lin_step(sun, size, maxsize);
        //s    *= 1.0-sstep(sun, 0.004, 0.0059)*0.5;

    float limb = 1.0 - pow3(lin_step(sun, 0.0, maxsize))*0.8;
        s    *= limb;

    #ifdef sun_transmittance_hq
        return s*sun_illum*15e2;
    #else
        return s*light_color[0]*15e2;
    #endif
}

vec3 get_moon(vec3 viewvec, vec3 albedo) {
    vec3 v      = -viewvec;
    vec3 sv     = normalize(moonvecView+v);
    float sun   = dot(sv, v);

    float s   = 1.0-lin_step(sun, 0.03, 0.08);

    return albedo*s*light_color[3]*6.0;
}

#include "/lib/frag/noise.glsl"

vec2 rotate_coord(vec2 pos, const float angle) {
    return vec2(cos(angle)*pos.x + sin(angle)*pos.y, 
                cos(angle)*pos.y - sin(angle)*pos.x);
}

vec3 get_stars(vec3 spos, vec3 svec) {
    vec3 plane  = svec/(svec.y+length(svec.xz)*0.66);
    float rot   = worldTime*rcp(2400.0);
    plane.x    += rot*0.6;
    plane.yz    = rotate_coord(plane.yz, (25.0/180.0)*pi);
    vec2 uv1    = floor((plane.xz)*768)/768;
    vec2 uv2    = (plane.xz)*0.04;

    vec3 starcol = vec3(0.5, 0.68, 1.0);
        starcol  = mix(starcol, vec3(1.0, 0.7, 0.6), noise_2d(uv2).x);
        starcol  = normalize(starcol)*(noise_2d(uv2*1.5).x+1.0);

    float star  = 1.0;
        star   *= noise_2d(uv1).x;
        star   *= noise_2d(uv1+0.1).x;
        star   *= noise_2d(uv1+0.26).x;

    star        = max(star-0.25, 0.0);
    star        = saturate(star*4.0);

    return star*starcol*0.1*sqrt(daytime.w);
}

#include "/lib/frag/bluenoise.glsl"

/* ------ cirrus clouds ------ */

float fbm(vec3 pos, vec3 offset, const float persistence, const float scale, const int octaves) {
    float n     = 0.0;
    float a     = 1.0;
    vec3 shift  = offset;

    for (int i = 0; i<octaves; ++i) {
        n      += value_3d(pos + shift)*a;
        pos.xz  = rotate_coord(pos.xz, pi*0.33);
        pos    *= scale;
        a      *= persistence;
    }

    return n;
}


const float pc_size     = 0.00026;
float vcloud_tick       = frameTimeCounter*0.1;

const float pc_highedge     = pcloud_alt + pcloud_depth;

float cloud_mie_dumb(float cos_theta, float g) {
    float sqG   = pow2(g);
    float a     = (1.0-sqG) / (2.0 + sqG);
    float b     = (1.0 + pow2(cos_theta)) / (1.0 + sqG - 2.0*g*cos_theta);

    return max(1.5 * a*b + g*cos_theta, 0.0)*rcp(pi);
}
float cloud_mie(float x, float g) {
    float mie   = 1.0 + pow2(g) - 2.0*g*x;
        mie     = (1.0 - pow2(g)) / ((4.0*pi) * mie*(mie*0.5+0.5));
    return mie;
}

float cloud_phase_cirrus(float cos_theta, float g) {
    float a     = cloud_mie(cos_theta, 0.44*g)+cloud_mie_dumb(cos_theta, 0.89*g)*0.75;
    float b     = cloud_mie(cos_theta, -0.25*g)*1.44;

    return mix(a, b, 0.38);
}

#ifdef pcloud_enabled

//classic cirrus
float pc_coverage(in vec3 pos) {
    float erode_low     = 1.0-sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.15);
    float erode_high    = sstep(pos.y, pcloud_alt+pcloud_depth*0.3, pc_highedge);
    float fade_low      = sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.1);
    float fade_high     = 1.0-sstep(pos.y, pcloud_alt+pcloud_depth*0.7, pc_highedge);

    float cov           = -0.04*daytime.x + 0.07*daytime.y + 0.01*daytime.z - 0.05*daytime.w;

    vec3 wind       = vec3(vcloud_tick, 0.0, 0.0);

    pos            *= pc_size;
    pos.x          *= 0.33;
    pos            += wind*0.1;

    //pos.xz         += noise_2d(pos.xz * 0.01).b * 4.0;

    float shape     = noise_2d(pos.xz*0.028).b;
        pos.z      += shape * 4.0;
        shape      += noise_2d(pos.xz*0.0015 + shape*0.0019).x*0.5;
        //shape      += noise_2d(pos.xz*0.001 - shape*0.002).x*0.2;
        shape      -= 0.72+cov;

        shape      *= fade_low;
        shape      *= fade_high;
        shape      -= erode_low*0.15;
        shape      -= erode_high*0.25;

    return max(shape, 0.0);
}

float pc_shape(vec3 pos, float coverage) {
    vec3 wind       = vec3(vcloud_tick, 0.0, 0.0);

    float dfade     = 0.3+sqrt(sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.2))*0.1;
        dfade      += sstep(pos.y, pcloud_alt+pcloud_depth*0.4, pc_highedge-pcloud_depth*0.1)*1.8;

    pos            *= pc_size;
    pos            += wind*0.1;

    float shape     = coverage;

    float d1        = 0.0;
    float n1        = 0.0;

        //pos.z      -= shape * 4.0 - 2.0;

    vec3 pos1       = pos * vec3(1.0, 1.0, 0.14) + vec3(shape, shape*0.2, shape*0.6 - 0.3).xyz*4.0;

    vec3 s1         = noise_3d(pos1*0.9);

        n1         += s1.x;     d1 += 1.0;
        n1         += s1.y*0.3;     d1 += 0.3;

        n1         /= d1;
        
        shape      -= (1.0-n1)*0.07;

    vec3 p3         = pos * 90.0 * vec3(1.0, 1.0, 0.3) + shape*60.0;

        shape      -= fbm(p3*1.5, wind, 0.5, 2.0, 2)*0.02;

    return max(shape*dfade, 0.0);
}

/*
//some kind of stratocumuli i guess
float pc_coverage(in vec3 pos) {
    float erode_low     = 1.0-sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.15);
    float erode_high    = sstep(pos.y, pcloud_alt+pcloud_depth*0.3, pc_highedge);
    float fade_low      = sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.1);
    float fade_high     = 1.0-sstep(pos.y, pcloud_alt+pcloud_depth*0.7, pc_highedge);

    float cov           = -0.04*daytime.x + 0.07*daytime.y + 0.01*daytime.z - 0.05*daytime.w;

    vec3 wind       = vec3(vcloud_tick, 0.0, 0.0);

    pos            *= pc_size;
    //pos.x          *= 0.33;
    pos            += wind*0.1;

    float s1        = noise_2d(pos.xz*0.01).b;

    float shape     = s1;
        shape      -= (0.5-noise_2d(pos.xz*0.04 + shape*0.02 + wind.xz * 0.004).b*0.5);
        shape      += noise_2d(pos.xz*0.2 + shape *0.1 + wind.xz * 0.01).b*0.2 * (1.0 + pow2(s1) * 0.6);
        shape      -= 0.4;
        //shape       = cov;

        shape       = max(shape, 0.0);

        shape      *= fade_low;
        shape      *= fade_high;
        shape      -= erode_low*0.15;
        shape      -= erode_high*0.25;

    return max(shape, 0.0);
}

float pc_shape(vec3 pos, float coverage) {
    vec3 wind       = vec3(vcloud_tick, 0.0, 0.0);

    float dfade     = 0.3+sqrt(sstep(pos.y, pcloud_alt, pcloud_alt+pcloud_depth*0.2))*0.1;
        dfade      += sstep(pos.y, pcloud_alt+pcloud_depth*0.4, pc_highedge-pcloud_depth*0.1)*1.0;

    pos            *= pc_size;
    pos            += wind*0.1;

    float shape     = coverage;

    float d1        = 0.0;
    float n1        = 0.0;

    vec3 pos1       = pos + shape * 1.1;

    vec3 s1         = noise_3d(pos1*0.4);

        n1         += s1.x;     d1 += 1.0;
        n1         += s1.y*0.3;     d1 += 0.3;

        n1         /= d1;
        
        shape      -= (1.0-n1) * 0.08 * (1.0 + coverage);

    vec3 p3         = pos * 50.0 + shape*20.0;

        shape      -= fbm(p3*1.5, wind, 0.5, 2.0, 2)*0.01;

    return max(shape*dfade, 0.0);
}*/

float pc_directOD(in vec3 pos, const int steps) {
    vec3 dir    = cloud_lvec;

    float basestep = mix(150.0, 20.0, saturate(float(steps-1)/4.0));
    float exponent = mix(2.2, 1.1, saturate(float(steps-8)/8.0));

    float stepsize  = basestep;
    float prev_step = stepsize;

    float od = 0.0;
    for(int i = 0; i < steps; ++i, pos += dir * stepsize) {
        if(pos.y > pc_highedge || pos.y < pcloud_alt) continue;
        prev_step = stepsize;
        stepsize *= exponent;

        float coverage  = pc_coverage(pos);
        if(coverage <= 0.0) continue;
        float attenCoeff = pc_shape(pos, coverage);
        od += attenCoeff*prev_step;
    } 

    return od;
}
float pc_skyOD(in vec3 pos, const int steps) {
    vec3 dir    = vec3(0.0, 1.0, 0.0);

    float stepsize = (pcloud_depth / steps);
        stepsize  *= (1.0-lin_step(pos.y, pcloud_alt, pc_highedge))*0.9+0.1;

    float od = 0.0;
    for(int i = 0; i < steps; ++i, pos += dir * stepsize) {
        if(pos.y > pc_highedge || pos.y < pcloud_alt) break;
        float coverage  = pc_coverage(pos);
        if(coverage <= 0.0) continue;
        float attenCoeff = pc_shape(pos, coverage);
        od += attenCoeff*stepsize;
    } 

    return od;
}

void compute_pc(inout vec3 scenecol, in vec3 vvec, in vec3 wvec, in float layer_transmittance) {
    //float within    = sstep(eyeAltitude, vcloud_alt-100.0, vcloud_alt) * (1.0-sstep(eyeAltitude, c_highedge, vc_highedge+100.0));
    bool visible    = wvec.y>0.0;

    if (visible) {
        const float pc_midalt = pcloud_alt+pcloud_depth*0.2;

        bool is_below   = eyeAltitude<pc_midalt;

        vec3 s      = wvec*((pc_midalt-eyeAltitude)/wvec.y);

        if (wvec.y<0.0 && is_below || wvec.y>0.0 && !is_below) s = vec3(0.0);

        vec3 rpos    = s + cameraPosition;

        //float dither = dither_bluenoise();

        float rlength   = pcloud_depth*0.5;

        vec3 scatter    = vec3(0.0);
        float transmittance = 1.0;
        float fade      = 0.0;
        float fdist     = pcloud_clip+1.0;

        vec3 sunlight   = (worldTime>23000 || worldTime<12900) ? light_color[0] : light_color[3];
        vec3 skylight   = light_color[1]*0.5;

        float vdotl     = dot(vvec, cloud_lvecView);

        float pfade     = saturate(cloud_mie(vdotl, 0.65));

        const float sigma_a = 1.20;         //absorption coeff
        const float sigma_s = 0.007;         //scattering coeff, can technically be assumed to be sigma_t since the albedo is close to 1.0
        const float sigma_t = 0.007;         //extinction coeff, 0.05-0.12 for cumulus, 0.04-0.06 for stratus
/*    
        const float sigma_a = 1.00;         //absorption coeff
        const float sigma_s = 0.035;         //scattering coeff, can technically be assumed to be sigma_t since the albedo is close to 1.0
        const float sigma_t = 0.04;         //extinction coeff, 0.05-0.12 for cumulus, 0.04-0.06 for stratus
*/
        /* --- calculate clouds --- */

        float dist  = distance(rpos, cameraPosition);
        float dfade = saturate(dist/pcloud_clip);
        if (!((1.0-dfade)<0.01)) {

        rpos.xz    *= 1.0-sstep(distance(rpos.xz, cameraPosition.xz), 0.0, 300000.0)*0.5;

        rpos.xz     = rotate_coord(rpos.xz, pi*rcp(8.0));

        float coverage  = pc_coverage(rpos);
        if (!(coverage<0.01)) {

        float density   = pc_shape(rpos, coverage);
        if (!(density<=0.0)) {

        if (fdist>pcloud_clip) fdist = dist;

        fade    = lin_step(dfade, 0.85, 0.99);

        float extinction = density * sigma_t;
        float stept     = expf(-extinction*rlength);
        float integral  = (1.0 - stept) / sigma_t;

        vec3 result_s   = vec3(0.0);

        float directod  = pc_directOD(rpos, 3)*sigma_a;
        float skyod     = pc_skyOD(rpos, 2)*sigma_a;

        float powder    = 1.0 - expf(-density * 40.0);
        float dpowder   = mix(powder, 1.0, pfade);

        for (int j = 0; j<8; ++j) {
            float n     = float(j);

            float s_d   = sigma_s * pow(0.5, n);    //scatter derivate
            float t_d   = sigma_t * pow(0.5, n);    //transmittance/attentuation derivate
            float phase = cloud_phase_cirrus(vdotl, pow(0.5, n));  //phase derivate

            result_s.x += expf(-directod*t_d) * phase * dpowder * s_d;
            result_s.y += expf(-skyod*t_d) * powder * s_d;
        }

        scatter    += result_s * integral;

        transmittance *= stept;
        }
        }
        }

        fdist   = max(fdist, 0.0);

        scatter.x *= 2.0;

        vec3 color  = (sunlight*scatter.x) + (skylight*scatter.y);

        fade        = saturate(1.0-fade);

        vec3 air        = air_density(wvec, vec3(0.0), 3);

        vec3 atmosfade  = expf(-extinct_mat*vec3(fdist)*0.66);

        float skyfade   = expf(-fdist*1.85e-5)*fade;

        scenecol        = mix(scenecol, scenecol * transmittance + color, atmosfade*layer_transmittance*skyfade);
    }
}
#endif

vec2 sincos(float x) {
    return vec2(sin(x), cos(x));
}

vec3 project_sphere(vec2 coord) {
    coord  *= vec2(tau, pi);
    vec2 lon = sincos(coord.x) * sin(coord.y);
    return vec3(lon.x, cos(coord.y), lon.y);
}

uniform float cloud_bsfade;

/* ------ cloud reflections ------ */

#define LQ_PASS

#ifdef vcloud_enabled

#include "/lib/atmos/vcloud.glsl"

void compute_vcr(inout vec3 scenecol, in vec3 wvec, inout float ltransmittance) {
    bool visible    = wvec.y>0.0;

    if (visible) {
        const float vc_midalt = vcloud_alt+vcloud_depth*0.5;

        vec3 bs     = wvec*((vcloud_alt-64.0)/wvec.y);
        vec3 ts     = wvec*((vc_highedge-64.0)/wvec.y);

        vec3 spos   = bs;
        vec3 epos   = ts;

        float dither = dither_bluenoise_s();

        const int basestep = 30;

        const float bl  = vcloud_depth/basestep;
        float stepl     = length((epos-spos)/basestep);
        float stepcoeff = stepl/bl;
            stepcoeff   = 0.45+clamp(stepcoeff-1.1, 0.0, 4.0)*0.5;
            //stepcoeff   = mix(stepcoeff, 10.0, pow2(within));
        int steps       = int(basestep*stepcoeff);

        vec3 rstep  = (epos-spos)/steps;
        vec3 rpos   = rstep*dither + spos + vec3(cameraPosition.x, 64.0, cameraPosition.z);
        float rlength = length(rstep);

        vec3 scatter    = vec3(0.0);
        float transmittance = 1.0;
        float fade      = 0.0;
        float fdist     = vcloud_clip+1.0;

        vec3 sunlight   = (worldTime>23000 || worldTime<12900) ? light_color[0] : light_color[3];
        vec3 skylight   = light_color[1]*0.33;

        float vdotl     = dot(wvec, cloud_lvec);

        float pfade     = saturate(cloud_mie(vdotl, 0.65));

        const float sigma_a = 1.00;         //absorption coeff
        const float sigma_s = 0.10;         //scattering coeff, can technically be assumed to be sigma_t since the albedo is close to 1.0
        const float sigma_t = 0.10;         //extinction coeff, 0.05-0.12 for cumulus, 0.04-0.06 for stratus

        for (int i = 0; i<steps; ++i, rpos += rstep) {
            if (transmittance < 0.05) break;
            if (rpos.y < vcloud_alt || rpos.y > vc_highedge) continue;

            float dist  = distance(rpos, cameraPosition);
            float dfade = saturate(dist/vcloud_clip);
            if ((1.0-dfade)<0.01) continue;

            float coverage  = vc_coverage(rpos);
            if (coverage<0.01) continue;

            float density   = vc_shape(rpos, coverage);
            if (density<=0.0) continue;

            if (fdist>vcloud_clip) fdist = dist;
            else fdist = mix(fdist, dist, transmittance);

            fade    = lin_step(dfade, 0.75, 0.99);

            float extinction = density * sigma_t;
            float stept     = expf(-extinction*rlength);
            float integral  = (1.0 - stept) / sigma_t;

            vec3 result_s   = vec3(0.0);

            float directod  = vc_directOD(rpos, 5)*sigma_a;
            float skyod     = vc_skyOD(rpos, 2)*sigma_a;

            float powder    = 1.0 - expf(-density * vc_powder_K);
            float dpowder   = min(powder, 1.0-expf(-directod*2.0));    //because sun od based powder simply adds this little "extra", using min here gives a much more pleasurable result
                dpowder     = mix(dpowder, 1.0, pfade);

            for (int j = 0; j<4; ++j) {
                float n     = float(j);

                float s_d   = sigma_s * pow(0.5, n);    //scatter derivate
                float t_d   = sigma_t * pow(0.5, n);    //transmittance/attentuation derivate
                float phase = cloud_phase(vdotl, pow(0.5, n));  //phase derivate

                result_s.x += expf(-directod*t_d) * phase * dpowder * s_d;
                result_s.y += expf(-skyod*t_d) * powder * s_d;
            }

            scatter    += result_s * integral * transmittance;

            transmittance *= stept;
        }
        fdist   = max(fdist, 0.0);
        transmittance = lin_step(transmittance, 0.05, 1.0);
        scatter.x *= 2.4;

        vec3 color  = (sunlight*scatter.x) + (skylight*scatter.y);

        fade        = saturate(1.0-fade);
        //fade        = pow2(fade);

        vec3 air       = air_density(wvec, vec3(0.0), 2);

        vec3 atmosfade = expf(-extinct_mat*vec3(fdist));

        float skyfade = expf(-fdist*cloud_atmos_density);

        scenecol = mix(scenecol, scenecol*transmittance + color, atmosfade*skyfade*fade);

        ltransmittance *= transmittance;
    }
}
#endif

#ifdef pcloud_enabled
void compute_pcr(inout vec3 scenecol, in vec3 wvec, in float layer_transmittance) {
    bool visible    = wvec.y>0.0;

    if (visible) {
        const float pc_midalt = pcloud_alt+pcloud_depth*0.2;

        vec3 s      = wvec*((pc_midalt-64.0)/wvec.y);

        vec3 rpos    = s + cameraPosition;

        float rlength   = pcloud_depth*0.5;

        vec3 scatter    = vec3(0.0);
        float transmittance = 1.0;
        float fade      = 0.0;
        float fdist     = pcloud_clip+1.0;

        vec3 sunlight   = (worldTime>23000 || worldTime<12900) ? light_color[0] : light_color[3];
        vec3 skylight   = light_color[1]*0.5;

        float vdotl     = dot(wvec, cloud_lvec);

        float pfade     = saturate(cloud_mie(vdotl, 0.65));

        const float sigma_a = 1.20;         //absorption coeff
        const float sigma_s = 0.007;         //scattering coeff, can technically be assumed to be sigma_t since the albedo is close to 1.0
        const float sigma_t = 0.007;         //extinction coeff, 0.05-0.12 for cumulus, 0.04-0.06 for stratus

        /* --- calculate clouds --- */

        float dist  = distance(rpos, cameraPosition);
        float dfade = saturate(dist/pcloud_clip);
        if (!((1.0-dfade)<0.01)) {

        rpos.xz    *= 1.0-sstep(distance(rpos.xz, cameraPosition.xz), 0.0, 300000.0)*0.5;

        rpos.xz     = rotate_coord(rpos.xz, pi*rcp(8.0));

        float coverage  = pc_coverage(rpos);
        if (!(coverage<0.01)) {

        float density   = pc_shape(rpos, coverage);
        if (!(density<=0.0)) {

        if (fdist>pcloud_clip) fdist = dist;

        fade    = lin_step(dfade, 0.85, 0.99);

        float extinction = density * sigma_t;
        float stept     = expf(-extinction*rlength);
        float integral  = (1.0 - stept) / sigma_t;

        vec3 result_s   = vec3(0.0);

        float directod  = pc_directOD(rpos, 2)*sigma_a;
        float skyod     = pc_skyOD(rpos, 2)*sigma_a;

        float powder    = 1.0 - expf(-density * 40.0);
        float dpowder   = mix(powder, 1.0, pfade);

        for (int j = 0; j<8; ++j) {
            float n     = float(j);

            float s_d   = sigma_s * pow(0.5, n);    //scatter derivate
            float t_d   = sigma_t * pow(0.5, n);    //transmittance/attentuation derivate
            float phase = cloud_phase_cirrus(vdotl, pow(0.5, n));  //phase derivate

            result_s.x += expf(-directod*t_d) * phase * dpowder * s_d;
            result_s.y += expf(-skyod*t_d) * powder * s_d;
        }

        scatter    += result_s * integral;

        transmittance *= stept;
        }
        }
        }

        fdist   = max(fdist, 0.0);

        scatter.x *= 2.0;

        vec3 color  = (sunlight*scatter.x) + (skylight*scatter.y);

        fade        = saturate(1.0-fade);

        vec3 air        = air_density(wvec, vec3(0.0), 2);

        vec3 atmosfade  = expf(-extinct_mat*vec3(fdist)*0.66);

        float skyfade   = expf(-fdist*cloud_atmos_density)*fade;

        scenecol        = mix(scenecol, scenecol * transmittance + color, atmosfade*layer_transmittance*skyfade);
    }
}
#endif

/* ------ ambient occlusion and gi ------ */

float max_depth3x3(sampler2D depthtex, vec2 coord, vec2 px) {

    float tl    = texture(depthtex, coord + vec2(-px.x, -px.y)).x;
    float tc    = texture(depthtex, coord + vec2(0.0, -px.y)).x;
    float tr    = texture(depthtex, coord + vec2(px.x, -px.y)).x;
    float tmin  = max(tl, max(tc, tr));

    float ml    = texture(depthtex, coord + vec2(-px.x, 0.0)).x;
    float mc    = texture(depthtex, coord).x;
    float mr    = texture(depthtex, coord + vec2(px.x, 0.0)).x;
    float mmin  = max(ml, max(mc, mr));

    float bl    = texture(depthtex, coord + vec2(-px.x, px.y)).x;
    float bc    = texture(depthtex, coord + vec2(0.0, px.y)).x;
    float br    = texture(depthtex, coord + vec2(px.x, px.y)).x;
    float bmin  = max(bl, max(bc, br));

    return max(tmin, max(mmin, bmin));
}

#ifdef ambientOcclusion_enabled
    #include "/lib/light/ao.glsl"
#endif

#ifdef rsm_enabled
    const bool colortex7Clear   = false;

    uniform sampler2D colortex7;

    #include "/lib/util/srgb.glsl"

    #include "/lib/light/warp.glsl"

    vec3 get_shadowcoord(vec3 viewpos, out float cd) {
        vec3 pos    = viewpos;
            pos     = viewMAD(gbufferModelViewInverse, pos);
            pos    += vec3(0.08)*lightvec;
        float a    = length(pos);
            pos     = viewMAD(shadowModelView, pos);
            pos     = projMAD(shadowProjection, pos);
            pos.z  *= 0.2;
            pos.z  -= 0.0012*(saturate(a/256.0));

        vec3 p2     = pos;
            cd      = -projMAD(shadowProjectionInverse, p2).z;

        return pos*0.5+0.5;
    }

    #include "/lib/light/rsm.glsl"

    vec3 get_lbounced(vec3 viewpos, vec3 scenenormal, vec3 scenepos, vec2 lmap, inout bool sampled, bool reproject, bool offscreen) {
        float cd    = 0.0;
        vec3 shadowspace_pos = get_shadowcoord(viewpos, cd);

        vec3 gi     = compute_rsm(scenenormal, shadowspace_pos, scenepos, cd, lmap.y, sampled, reproject, offscreen);

        return gi;
    }

    vec2 taa_reproject(vec2 coord, float depth) {
        float hand  = float(depth > 0.56);
        vec4 pos    = vec4(coord, depth, 1.0)*2.0-1.0;
            pos     = gbufferProjectionInverse*pos;
            pos    /= pos.w;
            pos     = gbufferModelViewInverse*pos;

        vec4 ppos   = pos + vec4(cameraPosition-previousCameraPosition, 0.0)*hand;
            ppos    = gbufferPreviousModelView*ppos;
            ppos    = gbufferPreviousProjection*ppos;

        return (ppos.xy/ppos.w)*0.5+0.5;
    }
#endif

vec4 textureBilateral(sampler2D tex, sampler2D depth, vec2 coord, const int lod, float fdepth) {
    vec4 data   = vec4(0.0);
    float sum   = 0.0;
    ivec2 posD  = ivec2(coord*viewSize);
    ivec2 posT  = ivec2(coord*viewSize*rcp(float(lod)));
    vec3 zmult  = vec3((far*near)*2.0, far+near, far-near);
        fdepth  = depth_lin(fdepth);
    //ivec2 p     = (ivec2(gl_FragCoord.xy + frameCounter) % 2 - 1)*0;
    
    for (int i = -1; i<2; i++) {
        for (int j = -1; j<2; j++) {
            ivec2 tcDepth = posD + ivec2(i, j)*lod * 2;
            float dsample = depth_lin(texelFetch(depth, tcDepth, 0).x);
            float w     = abs(dsample-fdepth)*zmult.x<1.0 ? 1.0 : 1e-5;
            ivec2 ct    = posT + ivec2(i, j) * 2;
            data       += texelFetch(tex, ct, 0)*w;
            sum        += w;
        }
    }
    data *= rcp(sum);

    return data;
}

vec4 textureBilateral_filter(sampler2D tex, sampler2D depth, const int lod, float fdepth, vec2 coord) {
    vec4 data   = vec4(0.0);
    float sum   = 0.0;
    ivec2 posD  = ivec2(coord*viewSize * float(lod));
    ivec2 posT  = ivec2(coord*viewSize);
    vec3 zmult  = vec3((far*near)*2.0, far+near, far-near);
        fdepth  = depth_lin(fdepth);
    //ivec2 p     = (ivec2(gl_FragCoord.xy + frameCounter) % 2 - 1)*0;
    
    for (int i = -2; i<3; i++) {
        for (int j = -2; j<3; j++) {
            ivec2 tcDepth = posD + ivec2(i, j)*lod;
            float dsample = depth_lin(texelFetch(depth, tcDepth, 0).x);
            float w     = abs(dsample-fdepth)*zmult.z<1.0 ? 1.0 : 1e-5;
            ivec2 ct    = posT + ivec2(i, j);
            vec4 tex    = texelFetch(tex, ct, 0)*w;
            data       += tex;
            sum        += w;
        }
    }
    data *= rcp(sum);

    return data;
}
/*
vec4 filter_spatial_gi(sampler2D tex, sampler2D depth, vec2 coord) {
    vec2 d_pos  = coord * 2.0;
    vec2 c_pos  = coord;

    float center_depth  = texture(depth, d_pos).x;
        center_depth    = depth_lin(center_depth) * far;

    vec4 sumcol     = vec4(0.0);
    float sumweight = 0.0;

    const float depth_thresh    = 0.2;

    const int sigma     = 5;
    const float rsigma  = 0.5 / (sigma * sigma);

    for (int x = -sigma; x<=sigma; x++) {
        for (int y = -sigma; y<=sigma; y++) {
            vec2 delta_pos  = vec2(x, y) * pixelSize;
            float blur  = expf(-dot(delta_pos, delta_pos) * rsigma) * rsigma;

            vec2 depth_pos  = d_pos + delta_pos * 2.0;
            float curr_depth = depth_lin(texture(depth, d_pos).x) * far;
            float depth_weight = pow2(1.0 - sstep(distance(curr_depth, center_depth), depth_thresh / 4, depth_thresh));

            vec2 curr_pos   = c_pos + delta_pos;
            vec4 curr_col   = texture(tex, curr_pos);

            float weight    = depth_weight;
            if (curr_col.a > 0.15) {
                sumcol     += curr_col * weight;
                sumweight  += 1.0 * weight;
            }
        }
    }

    sumcol /= max(sumweight, 1e-16);

    return sumcol;
}*/

vec4 filter_spatial_gi(sampler2D tex, sampler2D depth, vec2 coord) {
    vec2 d_pos  = coord * 2.0;
    vec2 c_pos  = coord;

    vec4 center_col     = texture(tex, c_pos);

    float center_depth  = texture(depth, d_pos).x;
        center_depth    = depth_lin(center_depth) * far;

    vec4 sumcol     = vec4(0.0);
    float sumweight = 0.0;
    
    const float depth_thresh    = 0.35;
    const float depth_rsigma    = ircp((depth_thresh * depth_thresh) * 2.0);
    const float depth_rpi       = ircp(pi) * ircp(depth_thresh);

    const float color_thresh    = 0.25;
    const float color_rsigma    = ircp((color_thresh * color_thresh) * 2.0);
    const float color_rpi       = ircp(pi) * ircp(color_thresh);

    const int sigma     = 5;
    const float rsigma  = 0.5 / (sigma * sigma);

    for (int x = -sigma; x<=sigma; x++) {
        for (int y = -sigma; y<=sigma; y++) {
            vec2 delta_pos  = vec2(x, y) * pixelSize;

            float blur  = expf(-dot(delta_pos, delta_pos) * rsigma) * rsigma;

            vec2 depth_pos  = d_pos + delta_pos * 2.0;
            float curr_depth = depth_lin(texture(depth, d_pos).x) * far;
            float depth_weight = pow2(1.0 - sstep(distance(curr_depth, center_depth), depth_thresh / 4, depth_thresh));

            vec2 curr_pos   = c_pos + delta_pos;
            vec4 curr_col   = texture(tex, curr_pos);

            float weight    = depth_weight;

            if (depth_weight > 0.95) {
                vec3 delta_col  = curr_col.rgb - center_col.rgb;
                float color_weight = expf(-dot(delta_col, delta_col) * color_rsigma) * color_rpi * blur;
                weight  = color_weight;
            }

            sumcol     += curr_col * weight;
            sumweight  += 1.0 * weight;
        }
    }

    sumcol /= max(sumweight, 1e-16);

    return sumcol;
}

void main() {
    vec4 scenecol   = stex(colortex0);
    float scenedepth = stex(depthtex1).x;

    vec3 viewpos    = screen_viewspace(vec3(coord, scenedepth));
    vec3 viewvec    = normalize(viewpos);
    vec3 scenepos   = view_scenespace(viewpos);
    vec3 svec       = normalize(scenepos);

    if (!landMask(scenedepth)) {
        #ifdef sky_hq_lookup
            vec3 skycol   = textureBicubic(colortex5, projectSky(svec)).rgb;
        #else
            vec3 skycol   = texture(colortex5, projectSky(svec)).rgb;
        #endif

        const float cLOD    = sqrt(CLOUD_RENDER_LOD);
        
        #ifdef vcloud_enabled
            vec2 cloudcoord = (coord)*rcp(cLOD)+vec2(1.0-rcp(cLOD), 0.0);

            #ifdef vcloud_gauss
                vec4 tex3   = gauss9(colortex3, cloudcoord, 1.0);
                vec4 tex5   = gauss9(colortex5, cloudcoord, 1.0);
            #else
                vec4 tex3   = texture(colortex3, cloudcoord);
                vec4 tex5   = texture(colortex5, cloudcoord);
            #endif
        #endif


        #ifdef atmos_refract
            vec3 viewvec_s = svec;
                float noiseint  = 1.0 - sstep(sunvec.y, -0.04, 0.04);
                    noiseint   *= sqrt(daytime.x + daytime.z);

                if (noiseint > 0.0) {
                float noise     = texture(noisetex, vec2(0.0 + frameTimeCounter*0.004, svec.y * 3.0 + frameTimeCounter*0.01)).x*0.01-0.005;
                    noise      += texture(noisetex, vec2(0.0 - frameTimeCounter*0.002, svec.y * 7.0 - frameTimeCounter*0.03)).x*0.005-0.0025;
                    noise      *= noiseint;
                    viewvec_s.z += noise;
                    //viewvec_s.y += noise;
                }
                viewvec_s = normalize(viewvec_s);
                viewvec_s = mat3(gbufferModelView) * viewvec_s;

            vec3 sun    = get_sun(viewvec_s);
        #else
            vec3 sun    = get_sun(viewvec);
        #endif
        
        #ifdef sun_transmittance_hq

            #ifdef sky_hq_lookup
                vec3 air_transmittance = textureBicubic(colortex5, projectSky(svec)+vec2(0.0, 0.3)).rgb;
            #else
                vec3 air_transmittance = texture(colortex5, projectSky(svec)+vec2(0.0, 0.3)).rgb;
            #endif

            sun    *= air_transmittance;
        #else
            sun *= pow3(lin_step(svec.y, -0.02, 0.02));
        #endif

        vec3 moonstars = get_moon(viewvec, scenecol.rgb) + get_stars(scenepos, svec);
            moonstars *= sstep(svec.y, -0.04, 0.01);

        #ifdef vcloud_enabled
            scenecol.rgb = skycol+(sun + moonstars)*mix(1.0, tex3.a, sstep(svec.y, 0.02, 0.15));

            #ifdef pcloud_enabled
                if (tex3.a>0.01) compute_pc(scenecol.rgb, viewvec, svec, tex3.a);
            #endif

            scenecol.rgb = mix(scenecol.rgb, scenecol.rgb * tex3.a + tex3.rgb, tex5.rgb);
        #else
            scenecol.rgb = skycol+(sun + moonstars);

            #ifdef pcloud_enabled
                compute_pc(scenecol.rgb, viewvec, svec, 1.0);
            #endif
        #endif

        //if (isnan3(scenecol.rgb) || isinf3(scenecol.rgb)) scenecol.rgb = vec3(0.0, 0.0, 40.0);
    }

    vec4 return3    = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 return7    = vec4(0.0);

    #if (defined ambientOcclusion_enabled || defined rsm_enabled)
        vec2 ao_coord   = (coord)*2.0;

        float d     = max_depth3x3(depthtex1, ao_coord, pixelSize*sqrt(2.0));

        if (clamp(ao_coord, -0.003, 1.003) == ao_coord && landMask(d)) {
            vec2 coord      = saturate(ao_coord);
            float scenedepth = stex(depthtex1).x;

            vec4 tex1   = stex(colortex1);

            #ifdef ambientOcclusion_enabled
            float ao    = calculate_dbao(depthtex1, scenedepth, coord);
            return3.a   = ao;
            #else
            return3.a   = 1.0;
            #endif

            #ifdef rsm_enabled
                vec2 reprojected = taa_reproject(coord, scenedepth);
                //vec4 previous_sample = texture(colortex7, reprojected * 0.5);
                vec4 previous_sample    = filter_spatial_gi(colortex7, depthtex1, reprojected * 0.5);
                bool prev_sampled   = previous_sample.a > 0.1;
                bool offscreen  = false;
                if (saturate(reprojected) != reprojected) {
                    prev_sampled = false;
                    offscreen = true;
                }

                vec3 scenenormal = decodeNormal(decode2x8(tex1.x));
                vec2 scenelmap  = decode2x8(tex1.y);
                vec3 viewpos    = screen_viewspace(vec3(coord, scenedepth));
                vec3 scenepos   = view_scenespace(viewpos);

                float fdist     = length(scenepos);
                float fade      = (1.0 - lin_step(fdist, 80.0, 144.0));

                bool curr_sample = false;

                vec3 gi_current  = vec3(0.0);
                if (fade > 0.0) {
                    gi_current = get_lbounced(viewpos, scenenormal, scenepos, scenelmap, curr_sample, prev_sampled, offscreen) * 9.0;
                    bool hand = d < 0.56;
                    if (hand) gi_current *= 0.25;

                    vec3 gi_accumulated = previous_sample.rgb;

                    vec2 vel    = (coord-reprojected)/(pixelSize * sqrt(2.0));
                    float taa_weight = saturate(1.0-sqrt(length(vel))/15.0);

                    if (curr_sample) {

                        if (prev_sampled) gi_accumulated = mix(previous_sample.rgb, gi_current, 0.24 + taa_weight * 0.74);
                        else gi_accumulated = gi_current;

                    } else if (prev_sampled && !curr_sample) {

                        gi_accumulated = previous_sample.rgb * 0.994 * (taa_weight * 0.85 + 0.15);

                    } else {
                        gi_accumulated = gi_current;
                    }

                    return7.rgb = gi_accumulated;
                    return7.a   = float(prev_sampled || curr_sample);

                    return3.rgb = gi_accumulated.rgb * fade;
                    //return3.rgb = previous_sample.rgb;
                } else {
                    return3.rgb = vec3(0.0);
                }

                //if (max3(previous_sample.rgb) > 0.9) return3.rgb = vec3(2.0, 0.0, 0.0);
                //if (max3(previous_sample.rgb) < 0.01) return3.rgb = vec3(0.0, 2.0, 0.0);

                //scenecol.rgb    = gi_accumulated.rgb;
            #endif
        }
    #endif

    /* ------ sky reflection ------ */

    vec4 return5    = stex(colortex5);

    const float cLOD    = sqrt(CLOUD_RENDER_LOD);

    vec2 scalecoord  = (coord-vec2(1.0-rcp(cLOD), 0.0))*cLOD;

    if (clamp(scalecoord, -0.003, 1.003) == scalecoord) {
        return5 = vec4(0.0, 0.0, 0.0, 1.0);
    }

    vec2 skyref_coord = (coord-vec2(0.55))*SKYREF_LOD;

    if (clamp(skyref_coord, -0.003, 1.003) == skyref_coord) {
        vec2 coord  = saturate(skyref_coord);
        vec3 svec   = project_sphere(coord)*vec3(1.0, 1.0, -1.0);
            //svec    = mat3(gbufferModelView)*svec;
        return5.rgb = texture(colortex5, projectSky(svec)).rgb;

        #ifdef cloud_reflections_enabled
            float layer_transmittance = 1.0;

            #ifdef vcloud_enabled
                compute_vcr(return5.rgb, svec, layer_transmittance);
            #endif

            #ifdef pcloud_enabled
                compute_pcr(return5.rgb, svec, layer_transmittance);
            #endif

            return5.a   = layer_transmittance;
        #endif
    }
    /*
    IBL sample vectors

    East (1.0, 0.0, 0.0)

    1-vec3(1.0, 0.15, 0.0)
    2-vec3(1.0, 0.40, 0.0)
    3-6 with z offsets of |0.4|

    South (0.0, 0.0, 1.0)

    1-vec3(0.0, 0.15, 0.0)
    2-vec3(0.0, 0.40, 0.0)
    3-6 with x offsets of |0.4|

    West (-1.0, 0.0, 0.0)

    1-vec3(-1.0, 0.15, 0.0)
    2-vec3(-1.0, 0.40, 0.0)
    3-6 with z offsets of |0.4|
    */

    //scenecol.rgb += vec3(10.0, 0.0, 0.0) * (1.0-lin_step(distance(svec, normalize(vec3(-0.4, 0.15, 1.0))), 0.0, 0.1));

    /*DRAWBUFFERS:0357*/
    gl_FragData[0]  = make_drawbuffer(scenecol);
    gl_FragData[1]  = clamp_drawbuffer(return3);
    gl_FragData[2]  = clamp_drawbuffer(return5);
    gl_FragData[3]  = clamp_drawbuffer(return7);
}