#version 400 compatibility

/*
Copyright (C) 2019 RRe36

All Rights Reserved unless otherwise explicitly stated.


By downloading this you have agreed to the license and terms of use.
These can be found inside the included license-file or here: https://rre36.github.io/kappa_shader_web/license/

Violating these terms may be penalized with actions according to the Digital Millennium Copyright Act (DMCA), the Information Society Directive and/or similar laws depending on your country.
*/

#include "/lib/common.glsl"

in vec2 coord;

flat in mat3x3 light_color;

uniform sampler2D colortex0;
uniform sampler2D colortex2;
uniform sampler2D colortex3;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

//uniform sampler3D depthtex2;
uniform sampler2D noisetex;

uniform int frameCounter;
uniform int isEyeInWater;

uniform float eyeAltitude;
uniform float frameTimeCounter;

uniform vec2 taaOffset;

uniform vec3 cameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

vec3 decode3x8(float a){
    int bf = int(a*65535.);
    return vec3(bf%32, (bf>>5)%64, bf>>11) / vec3(31,63,31);
}

vec3 screen_viewspace(vec3 screenpos, mat4 projInv) {
    screenpos   = screenpos*2.0-1.0;

    #ifdef taa_enabled
        screenpos.xy -= taaOffset;
    #endif

    vec3 viewpos    = vec3(vec2(projInv[0].x, projInv[1].y)*screenpos.xy + projInv[3].xy, projInv[3].z);
        viewpos    /= projInv[2].w*screenpos.z + projInv[3].w;
    
    return viewpos;
}

vec3 screen_viewspace(vec3 screenpos) {
    return screen_viewspace(screenpos, gbufferProjectionInverse);
}

vec3 view_scenespace(vec3 viewpos, mat4 mvInv) {
    return viewMAD(mvInv, viewpos);
}

vec3 view_scenespace(vec3 viewpos) {
    return view_scenespace(viewpos, gbufferModelViewInverse);
}

#include "/lib/frag/bluenoise.glsl"

#define noise_2d_only
#include "/lib/frag/noise.glsl"

float fog_tick  = frameTimeCounter*1.5;
const float fog_clip = 256.0;

float get_fog_density(vec3 pos) {
    vec3 wind   = vec3(fog_tick, -fog_tick*0.65, fog_tick*0.03);

    float dist  = distance(pos*vec3(1.0, 1.6, 1.0) - vec3(0.0, 96.0, 0.0), vec3(0.0));
        dist    = sstep(dist, 80.0, 256.0);

    pos    += wind*0.5;
    pos    *= 0.01;

    float delta = value_3d(pos*4.0 + wind*0.1)*2.0-1.0;
        delta += value_3d(pos*8.0 - wind*0.05)-0.5;

    pos    += delta * 0.7;

    float shape = value_3d(pos*8.0);
        shape  += value_3d(pos*16.0 - wind*0.1)*0.5;
        shape  -= 0.62;

        shape  -= dist * 0.38;
        shape  *= 1.0-dist*0.6;

    return max(shape*0.02, 0.0);
}

void compute_fog(out mat2x3 data, vec3 scenepos0, vec3 scenepos1, vec3 svec, bool sky) {
    //const int steps = 28;
    const float air_density = 3e4;

    float edist     = sky ? fog_clip : length(scenepos0);

    vec3 spos       = gbufferModelViewInverse[3].xyz;
    vec3 epos       = sky ? edist*svec : scenepos0;

    float dither    = dither_bluenoise();

    float bstep      = length((epos-spos));
    float stepcoeff  = saturate(bstep/256.0);
    int steps        = 8+int(stepcoeff*16*vfog_quality);

    vec3 rstep      = (epos-spos)/steps;
    vec3 rpos       = spos + rstep*dither;
    float rlength   = length(rstep);

    mat2x3 scatter  = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    float phaseIso  = 0.33*rcp(pi);

    vec3 skylight   = light_color[0];

    vec3 extinct_coeff = vec3(9.2e-6, 1.6e-5, 1.0e-5);
    vec3 scatter_coeff = vec3(3.5e-4);

    for (int i = 0; i<steps; ++i, rpos += rstep) {
        if (max(max(transmittance.x, transmittance.y), transmittance.z)<0.005) break;
        float density   = get_fog_density(rpos+cameraPosition);
        if (density <= 0.0) continue;

        float airmass = density * rlength * air_density;
        vec3 od     = extinct_coeff * airmass;

        vec3 stept  = saturate(exp(-od));
        vec3 integ  = saturate((stept - 1.0) / -od);
        vec3 vs     = transmittance * integ;

        vec3 sun_s  = scatter_coeff * (airmass * phaseIso) * vs;

        scatter[0]      += sun_s * transmittance;
        transmittance  *= stept;
    }

    vec3 color      = scatter[0]*skylight;

    data[0]         = color;
    data[1]         = transmittance;
}

vec3 blend_translucencies(vec3 scenecol, vec4 translucents, vec3 albedo) {
    vec3 color  = scenecol;
        color  *= mix(vec3(1.0), albedo, translucents.a);
        color   = mix(color, translucents.rgb, translucents.a);

    return color;
}

vec3 apply_fogdata(vec3 scenecol, mat2x3 data) {
    return scenecol.rgb*data[1] + data[0];
}

int decodeMatID16(float x) {
    return int(x*65535.0);
}

vec3 simple_fog(vec3 scenecolor, float d, vec3 color) {
    float dist      = max(0.0, d-32.0);
    float density   = 1.0-exp(-dist*1e-3);

    return mix(scenecolor, vec3(0.0), density);
}
vec3 water_fog(vec3 scenecolor, float d, vec3 color) {
    float dist      = max(0.0, d);
    float density   = dist*9.5e-1;
    vec3 scatter   = 1.0-exp(-min(density*0.8, 64e-1)*vec3(0.02, 0.24, 1.0));
    vec3 transmittance = exp(-density*vec3(1.0, 0.28, 0.06)*1.2);

    return scenecolor*transmittance + color*scatter*0.3;
}

void main() {
    vec4 scenecol   = stex(colortex0);  //that macro certainly makes it neater
    //vec4 tex1       = stex(colortex1);
    vec4 tex2       = stex(colortex2);
    vec4 tex3       = stex(colortex3);

    int mat_id      = decodeMatID16(tex2.x);

    float scenedepth0 = stex(depthtex0).r;
    float scenedepth1 = stex(depthtex1).r;

    vec3 viewpos0   = screen_viewspace(vec3(coord, scenedepth0));
    vec3 viewpos1   = screen_viewspace(vec3(coord, scenedepth1));

    vec3 scenepos0  = view_scenespace(viewpos0);
    vec3 scenepos1  = view_scenespace(viewpos1);

    vec3 sdir       = normalize(scenepos0);

    bool translucent = (scenedepth0<scenedepth1);

    bool water      = mat_id == 102;

    vec3 translucent_albedo = pow2(decode3x8(tex2.g));

    float dist0     = distance(scenepos0, gbufferModelViewInverse[3].xyz);
    float dist1     = distance(scenepos1, gbufferModelViewInverse[3].xyz);

    if (translucent && isEyeInWater==0){
        if (water) scenecol.rgb = water_fog(scenecol.rgb, dist1-dist0, light_color[1]);
    }

    scenecol.rgb    = blend_translucencies(scenecol.rgb, tex3, translucent_albedo);

    if (isEyeInWater==1) scenecol.rgb = water_fog(scenecol.rgb, dist0, light_color[1]);

    mat2x3 atmos;
        atmos[0] = vec3(0.0);
        atmos[1] = vec3(1.0);

    compute_fog(atmos, scenepos0, scenepos1, sdir, !landMask(scenedepth0));

    scenecol.rgb = apply_fogdata(scenecol.rgb, atmos);

    /*DRAWBUFFERS:0*/
    gl_FragData[0]  = make_drawbuffer(scenecol);
}